/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: CollectorServer.java mcarpentier $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.tools.webadmin.datacollector.server;

import java.sql.Connection;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.hibernate.HibernateException;
import org.hibernate.SQLQuery;
import org.hibernate.Session;
import org.hsqldb.Server;
import org.ow2.petals.tools.jmx.api.IPetalsAdminService.ContainerKeyNames;
import org.ow2.petals.tools.jmx.api.IPetalsAdminService.DomainKeyNames;
import org.ow2.petals.tools.jmx.api.exception.ConnectionErrorException;
import org.ow2.petals.tools.webadmin.datacollector.client.CollectorClient;
import org.ow2.petals.tools.webadmin.datacollector.client.exception.DataCollectorClientException;
import org.ow2.petals.tools.webadmin.datacollector.client.exception.DataCollectorClientInitializationException;
import org.ow2.petals.tools.webadmin.datacollector.client.service.AdminService;
import org.ow2.petals.tools.webadmin.datacollector.dao.DAOException;
import org.ow2.petals.tools.webadmin.datacollector.server.exception.DataCollectorClientDoesNotExistException;
import org.ow2.petals.tools.webadmin.datacollector.server.exception.DataCollectorServerException;
import org.ow2.petals.tools.webadmin.datacollector.server.exception.DataCollectorServerNotInitializedException;
import org.ow2.petals.tools.webadmin.datacollector.util.CollectorThread;
import org.ow2.petals.tools.webadmin.datacollector.util.HibernateHelper;
import org.ow2.petals.tools.webadmin.datacollector.util.Queuesmanager;
import org.ow2.petals.tools.webadmin.datacollector.util.ReconnectionThread;
import org.ow2.petals.tools.webadmin.datacollector.util.StringHelper;

public final class CollectorServer {

    private Map<String, CollectorClient> collectorClients;

    private static Server hsqlServer;

    private CollectorThread collectorThread;

    private ReconnectionThread reconnectionThread;

    private Queuesmanager qm;

    final static Logger log = Logger.getLogger(CollectorServer.class);

    /** properties given to the collector for initialization * */
    private String host;

    private String port;

    private String user;

    private String password;

    /**
     * 
     * @throws DataCollectorClientException
     * @throws DataCollectorClientInitializationException
     * @throws DataCollectorServerException
     * 
     */
    public CollectorServer(String host, String port, String user, String pwd)
            throws DataCollectorServerException {
        this.collectorClients = new HashMap<String, CollectorClient>();
        this.qm = new Queuesmanager();
        /**
         * first temp client, will try to connect to petals and the retrieve all
         * connected PEtALS servers conf
         */
        this.host = host;
        this.port = port;
        this.password = pwd;
        this.user = pwd;
        setupServer();
    }

    /**
     * do all server init operations
     * 
     * @param firstClient
     * @throws DataCollectorServerException
     */
    public void setupServer() throws DataCollectorServerException {
        try {
            final CollectorClient firstClient = new CollectorClient(this.host, this.port,
                    this.user, this.password);
            final AdminService adminService = firstClient.getAdminService();
            final Set<Map<String, String>> serversConfigurations = adminService
                    .getAllPetalsServersConfiguration();
            String domainName = "unknown";
            String domainMode = "unknown";
            String multicastIP = "unknown";
            String multicastPort = "unknown";
            for (final Map<String, String> petalsServer : serversConfigurations) {
                if (petalsServer.containsKey((DomainKeyNames.CONF_DOMAIN_NAME))) {
                    domainName = adminService.getContainerProperty(petalsServer,
                            DomainKeyNames.CONF_DOMAIN_NAME);
                    domainMode = adminService.getContainerProperty(petalsServer,
                            DomainKeyNames.CONF_DOMAIN_MODE);
                    multicastIP = adminService.getContainerProperty(petalsServer,
                            DomainKeyNames.CONF_NETWORK_MULTICAST_IP);
                    multicastPort = adminService.getContainerProperty(petalsServer,
                            DomainKeyNames.CONF_NETWORK_MULTICAST_PORT);
                } else {
                    /**
                     * PEtALS server found, build a {@link CollectorClient} for
                     * each server *
                     */

                    String host = adminService.getContainerProperty(petalsServer,
                            ContainerKeyNames.CONF_HOST);
                    String port = adminService.getContainerProperty(petalsServer,
                            ContainerKeyNames.CONF_JMX_JNDI_PORT);
                    String user = adminService.getContainerProperty(petalsServer,
                            ContainerKeyNames.CONF_USER);
                    String pwd = adminService.getContainerProperty(petalsServer,
                            ContainerKeyNames.CONF_PWD);

                    String name = adminService.getContainerProperty(petalsServer,
                            ContainerKeyNames.CONF_CONTAINER_NAME);

                    /** PEtALS version not supported* */
                    if ((StringHelper.isNullOrEmpty(name)) || (StringHelper.isNullOrEmpty(port))) {
                        throw new DataCollectorServerException(
                                "Can't find PEtALS server name/jmx port : PEtALS version may be not supported");
                    }

                    final CollectorClient cc = new CollectorClient(host, port, user, pwd);

                    /** set some server properties * */
                    /**
                     * we don't directly copy then whole hasmap as we may not
                     * want all properties to be transmitted *
                     */
                    cc.addServerProperty(ContainerKeyNames.CONF_CONTAINER_NAME, adminService
                            .getContainerProperty(petalsServer,
                                    ContainerKeyNames.CONF_CONTAINER_NAME));
                    cc.addServerProperty(ContainerKeyNames.CONF_SUBDOMAIN_NAME, adminService
                            .getContainerProperty(petalsServer,
                                    ContainerKeyNames.CONF_SUBDOMAIN_NAME));
                    cc.addServerProperty(ContainerKeyNames.CONF_STATE, adminService
                            .getContainerProperty(petalsServer, ContainerKeyNames.CONF_STATE));
                    cc.addServerProperty(ContainerKeyNames.CONF_HOST, adminService
                            .getContainerProperty(petalsServer, ContainerKeyNames.CONF_HOST));
                    cc.addServerProperty(ContainerKeyNames.CONF_JMX_JNDI_PORT, adminService
                            .getContainerProperty(petalsServer,
                                    ContainerKeyNames.CONF_JMX_JNDI_PORT));
                    cc.addServerProperty(ContainerKeyNames.CONF_NETWORK_BIND_PORT, adminService
                            .getContainerProperty(petalsServer,
                                    ContainerKeyNames.CONF_NETWORK_BIND_PORT));
                    cc.addServerProperty(ContainerKeyNames.CONF_DREAM_TCP_PORT, adminService
                            .getContainerProperty(petalsServer,
                                    ContainerKeyNames.CONF_DREAM_TCP_PORT));
                    cc.addServerProperty(ContainerKeyNames.CONF_JORAM_DOMAIN_PORT, adminService
                            .getContainerProperty(petalsServer,
                                    ContainerKeyNames.CONF_JORAM_DOMAIN_PORT));
                    cc.addServerProperty(ContainerKeyNames.CONF_JMX_RMI_PORT,
                            adminService.getContainerProperty(petalsServer,
                                    ContainerKeyNames.CONF_JMX_RMI_PORT));
                    cc.addServerProperty(ContainerKeyNames.CONF_JORAM_TCP_PORT, adminService
                            .getContainerProperty(petalsServer,
                                    ContainerKeyNames.CONF_JORAM_TCP_PORT));

                    this.collectorClients.put(cc
                            .getServerProperty(ContainerKeyNames.CONF_CONTAINER_NAME), cc);

                    /**
                     * if monitoring for this server is activated, start
                     * listening to messages*
                     */
                    if (cc.getMonitoringService().isMonitoring()) {
                        cc.getMonitoringService().startListening();
                    }
                }
            }
            for (CollectorClient c : this.collectorClients.values()) {
                c.addServerProperty(DomainKeyNames.CONF_DOMAIN_NAME, domainName);
                c.addServerProperty(DomainKeyNames.CONF_DOMAIN_MODE, domainMode);
                c.addServerProperty(DomainKeyNames.CONF_NETWORK_MULTICAST_IP, multicastIP);
                c.addServerProperty(DomainKeyNames.CONF_NETWORK_MULTICAST_PORT, multicastPort);
            }
            /** queues listening* */
            this.startQueuesListening();


        } catch (final DataCollectorClientInitializationException e) {
            throw new DataCollectorServerException(e);
        } catch (final DataCollectorClientException e) {
            throw new DataCollectorServerException(e);
        } catch (NumberFormatException e) {
            throw new DataCollectorServerException("Can't init JMX connection",e);
        }
    }

    /**
     * Check JDBC connection
     * 
     * @throws DataCollectorServerException
     */
    public void checkHSQLDBServerIsRunning() throws DataCollectorServerException {
        Session session = null;
        try {
            session = HibernateHelper.getSession();
            final Connection con = session.connection();
            log.debug("Connection to HSQLDB server OK : " + con.toString());
        } catch (final HibernateException e) {
            if (session != null) {
                session.close();
            }
            throw new DataCollectorServerException("Can't open connection to HSQLDB Server", e);
        }

    }

    public CollectorClient getCollectorClient(String serverName)
            throws DataCollectorServerException, DataCollectorClientDoesNotExistException,
            DataCollectorServerNotInitializedException {
        if ((this.collectorClients == null) || (this.collectorClients.isEmpty())) {
            throw new DataCollectorServerNotInitializedException();
        } else {
            if (this.collectorClients.containsKey(serverName)) {
                return this.collectorClients.get(serverName);
            } else {
                throw new DataCollectorClientDoesNotExistException("Can't find " + serverName
                        + " JMX client");
            }
        }
    }

    public List<Map<String, String>> getAllServersProperties() throws DataCollectorServerException,
            DataCollectorServerNotInitializedException {
        if ((this.collectorClients == null) || (this.collectorClients.isEmpty())) {
            throw new DataCollectorServerNotInitializedException();
        } else {
            final List<Map<String, String>> result = new ArrayList<Map<String, String>>();
            for (final String client : this.collectorClients.keySet()) {
                result.add(this.collectorClients.get(client).getServerProperties());
            }
            return result;
        }
    }

    public CollectorClient getDefaultCollectorClient() throws DataCollectorServerException,
            DataCollectorServerNotInitializedException {
        CollectorClient defaultClient = null;
        if ((this.collectorClients == null) || (this.collectorClients.isEmpty())) {

            throw new DataCollectorServerNotInitializedException();
        } else {
            final Iterator<CollectorClient> it = this.collectorClients.values().iterator();
            defaultClient = it.next();
        }
        return defaultClient;

    }

    public static void startHSQLServer() throws DataCollectorServerException {
        try {
            if (hsqlServer == null) {
                hsqlServer = (Server) Class.forName("org.hsqldb.Server").newInstance();
                hsqlServer.setDatabasePath(0, "localhost/datacollector-database");
                hsqlServer.setDatabaseName(0, "datacollector-database");
                hsqlServer.setSilent(true);
                hsqlServer.setTrace(false);
            }
            hsqlServer.checkRunning(true);
        } catch (final RuntimeException re) {
            /** server is not running, start it* */
            hsqlServer.start();
        } catch (InstantiationException e) {
            throw new DataCollectorServerException("Can't start HSQLDB Server", e);
        } catch (IllegalAccessException e) {
            throw new DataCollectorServerException("Can't start HSQLDB Server", e);
        } catch (ClassNotFoundException e) {
            throw new DataCollectorServerException("Can't start HSQLDB Server", e);

        }

    }

    public void startQueuesListening() {
        this.stopQueuesListening();
        /** TODO : implement this in a proper way * */
        this.collectorThread = new CollectorThread(this, 10000);
        this.collectorThread.start();
    }

    public static void stopHSQLServer() {
        if (hsqlServer != null) {
            Session session = null;
            try {
                session = HibernateHelper.getSession();
                SQLQuery q = session.createSQLQuery("SHUTDOWN IMMEDIATELY");
                hsqlServer.stop();
                hsqlServer.shutdown();
            } catch (final Exception e) {
                log.error("Can't shutdown database", e);
            } finally {
                if (session != null)
                    session.close();
            }

        }
    }

    public void stopQueuesListening() {
        if ((this.collectorThread != null) && (this.collectorThread.isAlive())) {
            this.collectorThread.interrupt();
        }
    }

    public void collectAllQueues() {
        String currentClient = "";
        try {
            if (this.collectorClients != null) {
                for (String client : this.collectorClients.keySet()) {
                    currentClient = client;
                    if (qm != null) {
                        qm.collectQueues(client, this.collectorClients.get(client)
                                .getTransportMonitoringService());
                    }
                }
            }
        } catch (DataCollectorClientException e) {
            CollectorServer.log.error("Can't retrieve queues status", e);
            if (e.getCause() instanceof ConnectionErrorException) {
                /** check client connection timeout* */
                
                if ((StringHelper.isNotNullAndNotEmpty(currentClient))
                        && (this.collectorClients.containsKey(currentClient))) {
                    CollectorClient cc = this.collectorClients.get(currentClient);
                    if (!cc.checkConnectionFailedCount()) {
                        this.collectorClients.remove(currentClient);
                        if ((this.collectorClients == null) || (this.collectorClients.isEmpty())) {
                            log.info("PEtALS connection seems to be lost, reconnection timer starting");
                            this.reconnectionThread = new ReconnectionThread(this, 10000);
                            this.reconnectionThread.start();
                        }
                    }
                }
            }
        } catch (DAOException e) {
            CollectorServer.log.error("Can't retreive queues status", e);
        }

    }
}
